@@ -44,15 +44,17 @@
 
 #include "xmpp.h"
 
-#include <qtextstream.h>
-#include <qpointer.h>
-#include <qtimer.h>
+#include <QTextStream>
+#include <QPointer>
+#include <QTimer>
 #include <QList>
 #include <QByteArray>
-#include <stdlib.h>
-#include "bytestream.h"
 #include <QtCrypto>
 #include <QUrl>
+//#include <stdio.h>
+#include <stdlib.h>
+
+#include "bytestream.h"
 #include "simplesasl.h"
 #include "securestream.h"
 #include "protocol.h"
@@ -65,6 +67,7 @@
 #include "td.h"
 #endif
 
+
 //#define XMPP_DEBUG
 
 using namespace XMPP;
@@ -175,6 +178,7 @@ public:
 		lang = "";
 
 		in_rrsig = false;
+		quiet_reconnection = false;
 
 		reset();
 	}
@@ -200,6 +204,7 @@ public:
 	QString connectHost;
 	int minimumSSF, maximumSSF;
 	QString sasl_mech;
+	QMap<QString,QString> mechProviders; // mech to provider map
 	bool doBinding;
 
 	bool in_rrsig;
@@ -236,6 +241,7 @@ public:
 	QTimer timeout_timer;
 	QTimer noopTimer;
 	int noop_time;
+	bool quiet_reconnection;
 };
 
 ClientStream::ClientStream(Connector *conn, TLSHandler *tlsHandler, QObject *parent)
@@ -288,13 +294,17 @@ ClientStream::ClientStream(const QString &host, const QString &defRealm, ByteStr
 
 ClientStream::~ClientStream()
 {
+	//fprintf(stderr, "\tClientStream::~ClientStream\n");
+	//fflush(stderr);
 	reset();
 	delete d;
-	//fprintf(stderr, "\tClientStream::~ClientStream\n");
 }
 
 void ClientStream::reset(bool all)
 {
+	//fprintf(stderr, "\tClientStream::reset\n");
+	//fflush(stderr);
+
 	d->reset();
 	d->noopTimer.stop();
 
@@ -306,6 +316,17 @@ void ClientStream::reset(bool all)
 	delete d->sasl;
 	d->sasl = 0;
 
+	if(all) {
+		while (!d->in.isEmpty()) {
+			delete d->in.takeFirst();
+		}
+	} else {
+		QSharedPointer<QDomDocument> sd;
+		foreach (Stanza *s, d->in) {
+			sd = s->unboundDocument(sd);
+		}
+	}
+
 	// client
 	if(d->mode == Client) {
 		// reset tls
@@ -335,12 +356,6 @@ void ClientStream::reset(bool all)
 
 		d->srv.reset();
 	}
-
-	if(all) {
-		while (!d->in.isEmpty()) {
-			delete d->in.takeFirst();
-		}
-	}
 }
 
 Jid ClientStream::jid() const
@@ -437,6 +452,24 @@ void ClientStream::continueAfterParams()
 	}
 }
 
+void ClientStream::setSaslMechanismProvider(const QString &m, const QString &p)
+{
+	d->mechProviders.insert(m, p);
+}
+
+QString ClientStream::saslMechanismProvider(const QString &m) const
+{
+	return d->mechProviders.value(m);
+}
+
+QCA::Provider::Context *ClientStream::currentSASLContext() const
+{
+	if (d->sasl) {
+		return d->sasl->context();
+	}
+	return 0;
+}
+
 void ClientStream::setSCRAMStoredSaltedHash(const QString &s) {
 	QCA::SASLContext *context = (QCA::SASLContext *)(d->sasl->context());
 	if (context) {
@@ -583,10 +616,10 @@ Stanza ClientStream::read()
 	}
 }
 
-void ClientStream::write(const Stanza &s, bool notify)
+void ClientStream::write(const Stanza &s)
 {
 	if(d->state == Active) {
-		d->client.sendStanza(s.element(), notify);
+		d->client.sendStanza(s.element());
 		processNext();
 	}
 }
@@ -625,7 +658,8 @@ void ClientStream::cr_connected()
 	d->client.doBinding = d->doBinding;*/
 
 	QPointer<QObject> self = this;
-	emit connected();
+	if (!d->quiet_reconnection)
+		emit connected();
 	if(!self)
 		return;
 
@@ -700,7 +734,8 @@ void ClientStream::ss_bytesWritten(qint64 bytes)
 void ClientStream::ss_tlsHandshaken()
 {
 	QPointer<QObject> self = this;
-	securityLayerActivated(LayerTLS);
+	if (!d->quiet_reconnection)
+		securityLayerActivated(LayerTLS);
 	if(!self)
 		return;
 	d->client.setAllowPlain(d->allowPlain == AllowPlain || d->allowPlain == AllowPlainOverTLS);
@@ -925,7 +960,8 @@ void ClientStream::srvProcessNext()
 void ClientStream::doReadyRead()
 {
 	//QGuardedPtr<QObject> self = this;
-	readyRead();
+	if (isActive())
+		emit readyRead();
 	//if(!self)
 	//	return;
 	//d->in_rrsig = false;
@@ -986,6 +1022,9 @@ void ClientStream::processNext()
 			//if(!d->in_rrsig && !d->in.isEmpty()) {
 			if(!d->in.isEmpty()) {
 				//d->in_rrsig = true;
+				//fprintf(stderr, "\tClientStream::processNext() QTimer::singleShot\n");
+				//fflush(stderr);
+
 				QTimer::singleShot(0, this, SLOT(doReadyRead()));
 			}
 
@@ -1036,6 +1075,9 @@ void ClientStream::processNext()
 #ifdef XMPP_DEBUG
 				qDebug("Break (Features)\n");
 #endif
+				if (d->client.unhandledFeatures.count()) {
+					emit haveUnhandledFeatures();
+				}
 				if(!d->tls_warned && !d->using_tls && !d->client.features.tls_supported) {
 					d->tls_warned = true;
 					d->state = WaitTLS;
@@ -1058,7 +1100,8 @@ void ClientStream::processNext()
 				d->jid = d->client.jid();
 				d->state = Active;
 				setNoopTime(d->noop_time);
-				authenticated();
+				if (!d->quiet_reconnection)
+					authenticated();
 				if(!self)
 					return;
 				break;
@@ -1078,10 +1121,10 @@ void ClientStream::processNext()
 				// store the stanza for now, announce after processing all events
 				// TODO: add a method to the stanza to mark them handled.
 				Stanza s = createStanza(d->client.recvStanza());
-				unsigned long sm_id = d->client.getNewSMId();
 				if(s.isNull())
 					break;
-				if (s.kind() == Stanza::Presence || s.kind() == Stanza::IQ) d->client.markStanzaHandled(sm_id);
+				if (d->client.sm.isActive())
+					d->client.sm.markStanzaHandled();
 				d->in.append(new Stanza(s));
 				break;
 			}
@@ -1103,10 +1146,30 @@ void ClientStream::processNext()
 				return;
 			}
 			case CoreProtocol::EAck: {
+				int ack_cnt = d->client.sm.takeAckedCount();
 #ifdef XMPP_DEBUG
-				qDebug() << "Received ack response: " << d->client.getNotableStanzasAcked();
+				qDebug() << "Stream Management: [INF] Received ack amount: " << ack_cnt;
 #endif
-				emit stanzasAcked(d->client.getNotableStanzasAcked());
+				emit stanzasAcked(ack_cnt);
+				break;
+			}
+			case CoreProtocol::ESMConnTimeout: {
+#ifdef XMPP_DEBUG
+				qDebug() << "Stream Management: [INF] Connection timeout";
+#endif
+				reset();
+				if (d->client.sm.state().isResumption()) {
+					d->state = Connecting;
+					emit warning(WarnSMReconnection);
+					d->quiet_reconnection = true;
+					if (d->client.sm.state().isLocationValid())
+						d->conn->setOptHostPort(d->client.sm.state().resumption_location.host , d->client.sm.state().resumption_location.port);
+					d->conn->connectToServer(d->server);
+				} else {
+					d->quiet_reconnection = false;
+					emit connectionClosed();
+				}
+				return;
 			}
 		}
 	}
@@ -1162,7 +1225,21 @@ bool ClientStream::handleNeed()
 				QCA::setProviderPriority("simplesasl", 10);
 			}
 
-			d->sasl = new QCA::SASL();
+			QStringList ml;
+			if(!d->sasl_mech.isEmpty())
+				ml += d->sasl_mech;
+			else
+				ml = d->client.features.sasl_mechs;
+
+			QString saslProvider;
+			foreach (const QString &mech, d->mechProviders.keys()) {
+				if (ml.contains(mech)) {
+					saslProvider = d->mechProviders[mech];
+					break;
+				}
+			}
+
+			d->sasl = new QCA::SASL(0, saslProvider);
 			connect(d->sasl, SIGNAL(clientStarted(bool,QByteArray)), SLOT(sasl_clientFirstStep(bool,QByteArray)));
 			connect(d->sasl, SIGNAL(nextStep(QByteArray)), SLOT(sasl_nextStep(QByteArray)));
 			connect(d->sasl, SIGNAL(needParams(QCA::SASL::Params)), SLOT(sasl_needParams(QCA::SASL::Params)));
@@ -1187,12 +1264,6 @@ bool ClientStream::handleNeed()
 				auth_flags = (QCA::SASL::AuthFlags) (auth_flags | QCA::SASL::RequireMutualAuth);
 			d->sasl->setConstraints(auth_flags,d->minimumSSF,d->maximumSSF);
 
-			QStringList ml;
-			if(!d->sasl_mech.isEmpty())
-				ml += d->sasl_mech;
-			else
-				ml = d->client.features.sasl_mechs;
-
 #ifdef IRIS_SASLCONNECTHOST
 			d->sasl->startClient("xmpp", QUrl::toAce(d->connectHost), ml, QCA::SASL::AllowClientSendFirst);
 #else
@@ -1214,7 +1285,8 @@ bool ClientStream::handleNeed()
 			d->ss->setLayerSASL(d->sasl, d->client.spare);
 			if(d->sasl_ssf > 0) {
 				QPointer<QObject> self = this;
-				securityLayerActivated(LayerSASL);
+				if (!d->quiet_reconnection)
+					securityLayerActivated(LayerSASL);
 				if(!self)
 					return false;
 			}
@@ -1268,18 +1340,6 @@ void ClientStream::doNoop()
 	}
 }
 
-// SM stuff
-bool ClientStream::isStreamManagementActive() {
-	return d->client.isStreamManagementActive();
-}
-
-void ClientStream::ackLastMessageStanza() {
-	 d->client.markLastMessageStanzaAcked();
-#ifdef XMPP_DEBUG
-	 qDebug() << "StreamManagement: markLastMessageStanzaAcked";
-#endif
-}
-
 void ClientStream::writeDirect(const QString &s)
 {
 	if(d->state == Active) {
@@ -1413,12 +1473,14 @@ void ClientStream::handleError()
 	}
 }
 
-ClientStream::SMState ClientStream::getSMState() const {
-	return d->client.getSMState();
+bool ClientStream::isResumed() const
+{
+	return d->client.sm.isResumed();
 }
 
-void ClientStream::setSMState(ClientStream::SMState state) {
-	d->client.setSMState(state);
+void ClientStream::setSMEnabled(bool e)
+{
+	d->client.sm.state().setEnabled(e);
 }
 
 
@@ -1427,6 +1489,16 @@ QStringList ClientStream::hosts() const
 	return d->client.hosts;
 }
 
+const StreamFeatures &ClientStream::streamFeatures() const
+{
+	return d->client.features;
+}
+
+QList<QDomElement> ClientStream::unhandledFeatures() const
+{
+	return d->client.unhandledFeatures;
+}
+
 
 //----------------------------------------------------------------------------
 // Debug
